
#include <ctype.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <string.h>
#include "stdlib.h"
#include <avr/io.h>
#include <math.h>
#include "video.h"

// ------- define --------

#define WALL_CHAR 186
// #define BLOCK_CHAR  //--spter
#define LEFT_WALL_X  0
#define RIGHT_WALL_X 13

#define TAST_PORT 		PORTD
#define TAST_PIN  		PIND
#define TAST_DDR 		DDRD                                                                                                                                                                                           
#define TAST_BITMASK 	0b11100000
#define TAST_LEFT 		0b01000000
#define TAST_RIGHT 		0b00100000
#define TAST_TURN		0b10000000

#define TAST_LEFT_FLAG  0b00000001
#define TAST_RIGHT_FLAG 0b00000010
#define TAST_TURN_FLAG  0b00000100
#define LOST_FLAG  		0b00100000
#define RESET_FLAG      0b01000000
#define RUNNING_FLAG    0b10000000

#define NO_MSG 		0
#define INIT_MSG 	1
#define LOST_MSG 	2

#define UP		0               // Block orientation and types
#define DOWN	1
#define LEFT	2
#define RIGHT	3

#define ATYPE	0
#define BTYPE	1
#define CTYPE	2
#define DTYPE	3
#define ETYPE	4
#define FTYPE	5
#define GTYPE	6

typedef struct point_tt {
	uint8_t x;
	uint8_t y;
} point_t;

typedef struct block_tt {
	point_t pos;
	uint8_t orient;
	uint8_t type;
} block_t;

// GLOBAL -----------------

volatile uint8_t gu8Flags;
char gcaStr[21];

block_t prev, cur, next;
uint16_t score, rows, gu16rand;

const char MCA_NEXT[] PROGMEM = "Next";
const char MCA_SCORE[] PROGMEM = "Score";
const char MCA_ROWS[] PROGMEM = "Rows";
const char MCA_ATMEGA[] PROGMEM = "ATmega";
const char MCA_TETRIS[] PROGMEM = "TETRIS";
const char MCA_GAME[] PROGMEM = "GAME";
const char MCA_OVER[] PROGMEM = "OVER";
const char MCA_PRESS[] PROGMEM = "Press";
const char MCA_ANYKEY[] PROGMEM = "any key";


// FUNKTIONEN -----------------------------------------------


void draw_vertline(point_t a, point_t b, char sym) {      // berarbeiten
	uint16_t i;
	for(i=CHAR_COLS*a.y; i<=CHAR_COLS*b.y; i+=CHAR_COLS) {
		screen_buffer[i+a.x] = sym;	
	}
}

uint8_t set_block(block_t block, char sym) {
	uint16_t offset0x, offset1x, offset2x, offset0y, offset1y, offset2y;       // Offset 1. Block ist immer 0/0
	
	switch(block.type){
case ATYPE:
	switch(block.orient) {
	case UP: case DOWN:
		offset0x=-2;
		offset0y=0;
		offset1x=-1;
		offset1y=0;
		offset2x=1;
		offset2y=0;
	break;
	case LEFT: case RIGHT:
		offset0x=0;
		offset0y=-2;
		offset1x=0;
		offset1y=-1;
		offset2x=0;
		offset2y=1;
	break;
	}
break;
case BTYPE:
	switch(block.orient) {
	case UP:
		offset0x=-1;
		offset0y=-1;
		offset1x=-1;
		offset1y=0;
		offset2x=1;
		offset2y=0;
	break;
	case RIGHT:
		offset0x=1;
		offset0y=-1;
		offset1x=0;
		offset1y=-1;
		offset2x=0;
		offset2y=1;		
	break;
	case DOWN:
		offset0x=1;
		offset0y=1;
		offset1x=-1;
		offset1y=0;
		offset2x=1;
		offset2y=0;		
	break;
	case LEFT:
		offset0x=-1;
		offset0y=1;
		offset1x=0;
		offset1y=-1;
		offset2x=0;
		offset2y=1;		
	break;
	}
break;
case CTYPE:
	switch(block.orient) {
	case UP:
		offset0x=1;
		offset0y=-1;
		offset1x=-1;
		offset1y=0;
		offset2x=1;
		offset2y=0;
	break;
	case RIGHT:
		offset0x=1;
		offset0y=1;
		offset1x=0;
		offset1y=-1;
		offset2x=0;
		offset2y=1;		
	break;
	case DOWN:
		offset0x=-1;
		offset0y=1;
		offset1x=-1;
		offset1y=0;
		offset2x=1;
		offset2y=0;		
	break;
	case LEFT:
		offset0x=-1;
		offset0y=-1;
		offset1x=0;
		offset1y=-1;
		offset2x=0;
		offset2y=1;		
	break;
	}
break;
case DTYPE:
	offset0x=1;
	offset0y=0;
	offset1x=1;
	offset1y=1;
	offset2x=0;
	offset2y=1;	
break;
case ETYPE:
	switch(block.orient) {
	case UP: case DOWN:
		offset0x=-1;
		offset0y=0;
		offset1x=0;
		offset1y=-1;
		offset2x=1;
		offset2y=-1;
	break;
	case LEFT: case RIGHT:
		offset0x=0;
		offset0y=-1;
		offset1x=1;
		offset1y=0;
		offset2x=1;
		offset2y=1;
	break;
	}
break;
case FTYPE:
	switch(block.orient) {
	case UP:
		offset0x=-1;
		offset0y=0;
		offset1x=0;
		offset1y=-1;
		offset2x=1;
		offset2y=0;
	break;
	case RIGHT:
		offset0x=0;
		offset0y=-1;
		offset1x=1;
		offset1y=0;
		offset2x=0;
		offset2y=1;		
	break;
	case DOWN:
		offset0x=-1;
		offset0y=0;
		offset1x=1;
		offset1y=0;
		offset2x=0;
		offset2y=1;		
	break;
	case LEFT:
		offset0x=0;
		offset0y=-1;
		offset1x=-1;
		offset1y=0;
		offset2x=0;
		offset2y=1;		
	break;
	}	
break;
case GTYPE:
	switch(block.orient) {
	case UP: case DOWN:
		offset0x=-1;
		offset0y=-1;
		offset1x=0;
		offset1y=-1;
		offset2x=1;
		offset2y=0;
	break;
	case LEFT: case RIGHT:
		offset0x=1;
		offset0y=-1;
		offset1x=1;
		offset1y=0;
		offset2x=0;
		offset2y=1;
	break;
	}
break;
};

	if(sym) {																      // sym == 0 heit Block lschen
		if(screen_buffer[block.pos.x+CHAR_COLS*block.pos.y] != 0 || block.pos.y >= CHAR_ROWS)
			return 1;
		if(screen_buffer[block.pos.x+offset0x+CHAR_COLS*(block.pos.y+offset0y)] != 0 || block.pos.y+offset0y >= CHAR_ROWS)
			return 1;
		if(screen_buffer[block.pos.x+offset1x+CHAR_COLS*(block.pos.y+offset1y)] != 0 || block.pos.y+offset1y >= CHAR_ROWS)
			return 1;
		if(screen_buffer[block.pos.x+offset2x+CHAR_COLS*(block.pos.y+offset2y)] != 0 || block.pos.y+offset2y >= CHAR_ROWS)
			return 1;
	}
	screen_buffer[block.pos.x+CHAR_COLS*block.pos.y] = sym;
	screen_buffer[block.pos.x+offset0x+CHAR_COLS*(block.pos.y+offset0y)] = sym;
	screen_buffer[block.pos.x+offset1x+CHAR_COLS*(block.pos.y+offset1y)] = sym;
	screen_buffer[block.pos.x+offset2x+CHAR_COLS*(block.pos.y+offset2y)] = sym;
	return 0;
}


void init_game(uint8_t msgtype) {

	point_t field_upleft={LEFT_WALL_X, 0}, field_downleft={LEFT_WALL_X, CHAR_ROWS-1},
			field_upright={RIGHT_WALL_X, 0}, field_downright={RIGHT_WALL_X,CHAR_ROWS-1};

	score = 0;
	rows = 0;

	cur.pos.x = 7;
	cur.pos.y = 2;
	cur.orient = rand;
	cur.type = rand;

	gu8Flags = 0;
	TAST_DDR &= ~TAST_BITMASK;    // TAST_DDR auf Input
	TAST_PORT |= TAST_BITMASK;    // TAST_PORT Pullups ein
	
	video_clear();

    draw_vertline(field_upleft, field_downleft, WALL_CHAR);
	draw_vertline(field_upright, field_downright, WALL_CHAR);
	strcpy_P(gcaStr, MCA_NEXT);
	video_print(15, 2, gcaStr);
	strcpy_P(gcaStr, MCA_ROWS);
	video_print(15, 13, gcaStr);
	strcpy_P(gcaStr, MCA_SCORE);
	video_print(15, 19, gcaStr);
	utoa(rows, gcaStr, 5);
	video_print(15, 15, gcaStr);
	utoa(score, gcaStr, 5);
	video_print(15, 21, gcaStr);
	
	switch(msgtype) {
	case INIT_MSG:
		strcpy_P(gcaStr, MCA_ATMEGA);
		video_print(3, 8, gcaStr);
		strcpy_P(gcaStr, MCA_TETRIS);
		video_print(5, 10, gcaStr);		
		strcpy_P(gcaStr, MCA_PRESS);
		video_print(4, 13, gcaStr);
		strcpy_P(gcaStr, MCA_ANYKEY);
		video_print(3, 14, gcaStr);			
	break;
	case LOST_MSG:
		strcpy_P(gcaStr, MCA_GAME);
		video_print(3, 8, gcaStr);
		strcpy_P(gcaStr, MCA_OVER);
		video_print(6, 10, gcaStr);		
		strcpy_P(gcaStr, MCA_PRESS);
		video_print(4, 13, gcaStr);
		strcpy_P(gcaStr, MCA_ANYKEY);
		video_print(3, 14, gcaStr);		
	
	break;
	default:
	break;
	}
	
}

void clear_rows(block_t block) {
uint16_t x, y, y2;
uint16_t rows_cleared = 0, filled;
	for(y=block.pos.y-2; y<CHAR_ROWS; y++) {
		filled = 0;
		for(x=LEFT_WALL_X+1; x<RIGHT_WALL_X; x++) {
			if(screen_buffer[x+CHAR_COLS*y] == 'X')
				filled++;
		}
		if(filled == RIGHT_WALL_X - LEFT_WALL_X - 1) {
			rows_cleared++;
			for(x=LEFT_WALL_X+1; x<RIGHT_WALL_X; x++) {
				for(y2=y; y2>0; y2--)
					screen_buffer[x+CHAR_COLS*y2] = screen_buffer[x+CHAR_COLS*(y2-1)];
				screen_buffer[x] = 0;
				}
			}
	}
	rows += rows_cleared;
	switch(rows_cleared) {
	case 1:
		score += 1; 
	break;
	case 2:
		score += 3;
	break;
	case 3:
		score += 12;
	break;
	case 4:
		score += 25;
	break;
	default:
	break;
	}
	ultoa(rows, gcaStr, 5);
	video_print(15, 15, gcaStr);
	ultoa(score, gcaStr, 5);
	video_print(15, 21, gcaStr);
}
//------------------------------------------------------------------------------
//Hier startet das Hauptprogramm


int main (void) 
{	
	uint16_t u16TickNowFast, u16TickCntFast,
			 u16TickNowPoll, u16TickCntPoll,
			 u16TickNowSlow, u16TickCntSlow;
	
    u16TickNowSlow = u16TickCntSlow = 0;
    u16TickNowFast = u16TickCntFast = 0;
    u16TickNowPoll = u16TickCntPoll = 0;	
	
	video_init();
	gu8Flags |= RESET_FLAG;
	
    while(1)
	{
	// POLLING -----------------
	if( (u16TickCntPoll - u16TickNowPoll ) > 5000 )
		{
		if( !(TAST_PIN & TAST_LEFT))
			gu8Flags |= TAST_LEFT_FLAG;
			
		if( !(TAST_PIN & TAST_RIGHT))
			gu8Flags |= TAST_RIGHT_FLAG;
			
		if( !(TAST_PIN & TAST_TURN))
			gu8Flags |= TAST_TURN_FLAG;
			
		if( !(gu8Flags & RUNNING_FLAG)  && (!(TAST_PIN & TAST_TURN) || !(TAST_PIN & TAST_LEFT) || !(TAST_PIN & TAST_RIGHT))) {
			
			gu16rand = u16TickNowPoll;
			next.pos.x = 17;
			next.pos.y = 6;
			next.orient = gu16rand % 4;
			next.type = gu16rand % 7;
			set_block(next, 'X');

			init_game(0);
			gu8Flags |= RUNNING_FLAG;		
			}
			
		if(gu8Flags & RESET_FLAG) {
			init_game(INIT_MSG);
			gu8Flags &= ~RUNNING_FLAG;				
		}
		if(gu8Flags & LOST_FLAG) {
			init_game(LOST_MSG);
			gu8Flags &= ~RUNNING_FLAG;				
		}
			
		u16TickNowPoll = u16TickCntPoll;
		};
	
	// SLOW  -----------------
	if( (u16TickCntSlow - u16TickNowSlow ) > 40000 )
		{
		
		prev = cur;		
		set_block(cur, 0);

		cur.pos.y++;
		if(set_block(cur, 'X') != 0) {
			set_block(prev, 'X');

			clear_rows(prev);
	
			cur = next;
			cur.pos.x = 7;
			cur.pos.y = 2;
			
			set_block(next, 0);
			next.pos.x = 17;
			next.pos.y = 6;
			next.orient = u16TickNowFast % 4;
			next.type = u16TickNowFast % 7;
			set_block(next, 'X');
			if(set_block(cur, 'X') != 0) {
				gu8Flags |= LOST_FLAG;
				gu8Flags &= ~RUNNING_FLAG;
			}
		}
		else 
			set_block(cur, 'X');
	
		u16TickNowSlow = u16TickCntSlow;
		};
		
	// FAST  -----------------
	if( (u16TickCntFast - u16TickNowFast ) > 20000 )	
		{
		if(gu8Flags & TAST_LEFT_FLAG) {
			prev = cur;	
			set_block(cur, 0);
	
			gu8Flags &= ~TAST_LEFT_FLAG;

			cur.pos.x--;
			
			if(set_block(cur, 'X') != 0)
				cur = prev;
			set_block(cur, 'X');
		}
		if(gu8Flags & TAST_RIGHT_FLAG) {
		
			prev = cur;		
			set_block(cur, 0);
			
			gu8Flags &= ~TAST_RIGHT_FLAG;
	
			cur.pos.x++;
			
			if(set_block(cur, 'X') != 0)
				cur = prev;
			set_block(cur, 'X');
		}
		if(gu8Flags & TAST_TURN_FLAG) {
		
			prev = cur;		
			set_block(cur, 0);
			
			gu8Flags &= ~TAST_TURN_FLAG;
	
			switch(cur.orient) {
			case UP:
				cur.orient = RIGHT;
			break;
			case RIGHT:
				cur.orient = DOWN;
			break;
			case DOWN:
				cur.orient = LEFT;
			break;
			case LEFT:
				cur.orient = UP;
			break;			
			}
			
			if(set_block(cur, 'X') != 0)
				cur = prev;
			set_block(cur, 'X');
		}
		
		u16TickNowFast = u16TickCntFast;
		};

		u16TickCntPoll++;
		if(gu8Flags & RUNNING_FLAG) {
			u16TickCntSlow++;
			u16TickCntFast++;
		}
    };
	
}

